DAT programName byte "BotControl130226b", 0
CON
{{
  ******* Public Notes *******

  This program is used to read from input devices
  and set target speeds in the parent object.
  Developed with a Rover 5 chassis.
  
  
}}
{
  ******* Private Notes *******
  MecanumRc130220a modified from VexMecanumRc121219b130131a 
  21a Rename InputRc130221a.
  23a Rename InputRcNun130223a.
  23a Add Wii Nunchuck as an optional input device.
  24a Create array for channel centers and deadbands.
  24b Merge with Motion object.
  24b InputRcNun130224b to BotControl130224b
  24d Add tempChannel array to hold values
  prior final values stored.
  25a Start adding auto dectect feature.
  
}
CON


  RX_CHANNELS = Rc#POSSIBLE_CHANNELS
  HIGHEST_CHANNEL = RX_CHANNELS - 1

  I2C_CLOCK = 28       '**should this infomation be in the header? 
  I2C_DATA = 29

  
  

  
  ' channel names    **should this infomation be in the header? Different orders depending
                    ' depending on controller/robot?
  #0, THROTTLE_CHANNEL, X_CHANNEL, Y_CHANNEL, ROTATE_CHANNEL, GEAR_CHANNEL, AUX_CHANNEL
 
  ' type enumeration
  #0, NORMAL_RANGE, ZERO_CENTER_RANGE

  ' controllerType enumeration
  '#0, RC_CONTROLLER, NUNCHUCK_CONTROLLER

CON

  ' RC RX Constants
  DEAD_BAND = 11  ' All the same? What about throttle and landing gear?
  
CON

   'Nunchuck Constants  **should this infomation be in the header?
  JOYX_FAR_LEFT = 26
  JOYX_FAR_RIGHT = 224
  JOYX_LEFT_CENTER_EDGE = 125
  JOYX_RIGHT_CENTER_EDGE = 145 '135
  JOYY_FAR_FORWARD = 223
  JOYY_FAR_BACKWARD = 29
  JOYY_FORWARD_CENTER_EDGE = 140 '135
  JOYY_BACKWARD_CENTER_EDGE = 128
  CHANGE_DIGIT_THRESHOLD = 30
  OTHER_AXIS_LIMIT = 10
  SETTING_DIVIDER = 10
  
  AX_RAW_CENTER = 492
  AX_JITTER = 15 '5
  AY_RAW_CENTER = 500
  AY_JITTER = 5
  AZ_RAW_CENTER = 656
  AZ_BUFFER = 3
  
  UP_HIGH_THRESHOLD = 710
  UP_LOW_THRESHOLD = 640
  DOWN_HIGH_THRESHOLD = 310
  DOWN_LOW_THRESHOLD = 220
  MID_HIGH_THRESHOLD = 570
  MID_LOW_THRESHOLD = 420
  
CON ' from Motion object

  MAXSPEED = 1000
  R45 = pi / 4.0
  HIGHEST_WHEEL_ID = 3
  TOTAL_WHEELS = 4
  

  ' inputType enumeration
  #0, PC_ONLY_INPUT, WII_NUNCHUCK_INPUT, RC_INPUT, WII_AND_RC_INPUT, AUTO_DETECT_INPUT
  
  ' wheelMode enumeration  this needs to match header constants
  #0, INDEPENDENT_WHEELS, TREADED_WHEELS, MECANUM_WHEELS 

  
  CIRCLE = 360
  
  BUTTONS_UP_NUNCHUCK = 0 '-1
  
VAR

  long channelPtr', center[RX_CHANNELS]
  long deadBand[RX_CHANNELS]
  long deadEdgeLow[RX_CHANNELS], deadEdgeHigh[RX_CHANNELS]

  long tempTargetSpeed[TOTAL_WHEELS], fPseudoMultiplier
  long speedPtr, pinPtr, globalX, globalY, globalZ
  long tempChannel[RX_CHANNELS], pseudoCircle
  
  byte controllerType, wheelType

OBJ

  Rc : "RcReceiver121116c"
  Nunchuck  : "jm_nunchuk_ez_v3"
  F : "F32"
   
PUB Start(wheelType_, controllerType_, speedPtr_, channelPtr_, pinPtr_, pseudoMultiplier) 

  wheelType := wheelType_
  controllerType := controllerType_
  speedPtr := speedPtr_
  channelPtr := channelPtr_
  pinPtr := pinPtr_
  
  if controllerType == AUTO_DETECT_INPUT
    Nunchuck.init(I2C_CLOCK, I2C_DATA) 
    Nunchuck.scan 
    'waitcnt(clkfreq / 10 + cnt) 
    'result := Nunchuck.accx
    result := Nunchuck.accy
    if result <> $3FF
      InitNunchuck
      controllerType := WII_NUNCHUCK_INPUT
    else
      controllerType := RC_INPUT
      result := Rc.BuildMask(byte[pinPtr], byte[pinPtr][1], byte[pinPtr][2], {
      } byte[pinPtr][3], byte[pinPtr][4], byte[pinPtr][5])     
      Rc.Start(result)
      InitRc
  elseif controllerType == RC_INPUT
    result := Rc.BuildMask(byte[pinPtr], byte[pinPtr][1], byte[pinPtr][2], {
    } byte[pinPtr][3], byte[pinPtr][4], byte[pinPtr][5])     
    Rc.Start(result)
    InitRc
  elseif controllerType == WII_NUNCHUCK_INPUT
    Nunchuck.init(I2C_CLOCK, I2C_DATA) 
    InitNunchuck

  if wheelType == MECANUM_WHEELS
    InitF(pseudoMultiplier)
  result := @channelPtr
    
PUB BytePtr

  result := @controllerType
  
PUB InitF(pseudoMultiplier)

  result := F.Start
  fPseudoMultiplier := pseudoMultiplier
  fPseudoMultiplier := F.FFloat(fPseudoMultiplier)
  pseudoCircle := CIRCLE * pseudoMultiplier 

PUB RefreshInput

  if controllerType == RC_INPUT
  {  
    repeat result from 0 to HIGHEST_CHANNEL
      tempChannel[result] := Rc.GetRc(result)
      'long[channelPtr][result] := Rc.GetRc(result)
      if ||tempChannel[result] < deadBand[result]
        long[channelPtr][result] := 0
      else
        long[channelPtr][result] := tempChannel[result]
     }
    repeat result from 0 to HIGHEST_CHANNEL
      tempChannel[result] := Rc.GetRc(result)
      
      if tempChannel[result] < -deadBand[result]
        long[channelPtr][result] := tempChannel[result] + deadBand[result]
      elseif tempChannel[result] =< deadBand[result]
        long[channelPtr][result] := 0
      else  
        long[channelPtr][result] := tempChannel[result] - deadBand[result]
        
        
    globalX := long[channelPtr][X_CHANNEL] * 2
    globalY := long[channelPtr][Y_CHANNEL] * 2
    globalZ := long[channelPtr][ROTATE_CHANNEL] 
  elseif controllerType == WII_NUNCHUCK_INPUT
    Nunchuck.scan
    
    tempChannel[X_CHANNEL] := Nunchuck.joyx
    tempChannel[Y_CHANNEL] := Nunchuck.joyy
    tempChannel[ROTATE_CHANNEL] := Nunchuck.accx
    tempChannel[THROTTLE_CHANNEL] := Nunchuck.accy
    tempChannel[AUX_CHANNEL] := Nunchuck.accz
    tempChannel[GEAR_CHANNEL] := Nunchuck.buttons

    repeat result from 0 to HIGHEST_CHANNEL
      if result == GEAR_CHANNEL
        long[channelPtr][result] := tempChannel[result]
      else
        if tempChannel[result] < deadEdgeLow[result]
          long[channelPtr][result] := tempChannel[result] - deadEdgeLow[result]
        elseif tempChannel[result] =< deadEdgeHigh[result]
          long[channelPtr][result] := 0
        else  
          long[channelPtr][result] := tempChannel[result] - deadEdgeHigh[result]
        
    globalX := long[channelPtr][X_CHANNEL] * -10
    globalY := long[channelPtr][Y_CHANNEL] * 10
    globalZ := long[channelPtr][ROTATE_CHANNEL] * -4

 {
****elseif controllerType == WII_NUNCHUCK_INPUT
    Nunchuck.scan
    
    tempChannel[X_CHANNEL] := Nunchuck.joyx
    tempChannel[Y_CHANNEL] := Nunchuck.joyy
    tempChannel[ROTATE_CHANNEL] := Nunchuck.accx
    tempChannel[THROTTLE_CHANNEL] := Nunchuck.accy
    tempChannel[AUX_CHANNEL] := Nunchuck.accz
    tempChannel[GEAR_CHANNEL] := Nunchuck.buttons
    
    repeat result from 0 to HIGHEST_CHANNEL
      tempChannel[result] -= center[result]
    repeat result from 0 to HIGHEST_CHANNEL
      if ||tempChannel[result] < deadBand[result]
        long[channelPtr][result] := 0
      else                                    
        long[channelPtr][result] := tempChannel[result]  
    globalX := long[channelPtr][X_CHANNEL] * 10
    globalY := long[channelPtr][Y_CHANNEL] * 10
    globalZ := long[channelPtr][ROTATE_CHANNEL] * -4
    ***         }
PUB Refresh 

  RefreshInput
  if wheelType == INDEPENDENT_WHEELS or wheelType == TREADED_WHEELS
    SpeedFromXY
  elseif wheelType == MECANUM_WHEELS
    ComputeVector
    
PUB IncrementAngle(psuedoAngle, localPseudoIncrement)

  psuedoAngle += localPseudoIncrement
  if psuedoAngle => pseudoCircle
    psuedoAngle -= pseudoCircle
  elseif psuedoAngle < 0
    psuedoAngle += pseudoCircle

  result := psuedoAngle
     
PRI InitRc

  repeat result from 0 to HIGHEST_CHANNEL
    'center[result] := 0 '1500
    deadBand[result] := DEAD_BAND
  deadBand[GEAR_CHANNEL] := 0
   
PRI InitNunchuck


  'center[X_CHANNEL] :=  (JOYX_RIGHT_CENTER_EDGE + JOYX_LEFT_CENTER_EDGE) / 2
  deadBand[X_CHANNEL] := JOYX_RIGHT_CENTER_EDGE - JOYX_LEFT_CENTER_EDGE 
  deadEdgeLow[X_CHANNEL] := JOYX_LEFT_CENTER_EDGE
  deadEdgeHigh[X_CHANNEL] := JOYX_RIGHT_CENTER_EDGE
  'center[Y_CHANNEL] := (JOYY_FORWARD_CENTER_EDGE + JOYY_BACKWARD_CENTER_EDGE) / 2 
  deadBand[Y_CHANNEL] := JOYY_FORWARD_CENTER_EDGE - JOYY_BACKWARD_CENTER_EDGE
  deadEdgeLow[Y_CHANNEL] := JOYY_BACKWARD_CENTER_EDGE
  deadEdgeHigh[Y_CHANNEL] := JOYY_FORWARD_CENTER_EDGE
  'center[ROTATE_CHANNEL] := AX_RAW_CENTER
  deadBand[ROTATE_CHANNEL] := AX_JITTER
  deadEdgeLow[ROTATE_CHANNEL] := AX_RAW_CENTER - AX_JITTER
  deadEdgeHigh[ROTATE_CHANNEL] := AX_RAW_CENTER + AX_JITTER 
  'center[THROTTLE_CHANNEL] := AY_RAW_CENTER
  deadBand[THROTTLE_CHANNEL] := AY_JITTER
  deadEdgeLow[THROTTLE_CHANNEL] := AY_RAW_CENTER - AY_JITTER 
  deadEdgeHigh[THROTTLE_CHANNEL] := AY_RAW_CENTER + AY_JITTER 
  'center[GEAR_CHANNEL] := 1
  deadBand[GEAR_CHANNEL] := 0
  deadEdgeLow[GEAR_CHANNEL] := 1
  deadEdgeHigh[GEAR_CHANNEL] := 1
  'center[AUX_CHANNEL] := AZ_RAW_CENTER
  deadBand[AUX_CHANNEL] := AZ_BUFFER
  deadEdgeLow[AUX_CHANNEL] := AZ_RAW_CENTER - AZ_BUFFER
  deadEdgeHigh[AUX_CHANNEL] := AZ_RAW_CENTER + AZ_BUFFER
  
PUB GetFileName

  result := @programName

PUB SpeedFromXY'(globalX, globalY)
    
  'if true ' inputType == RC_INPUT
  long[speedPtr][2] := (globalY - globalX) '* 2
  long[speedPtr][0] := long[speedPtr][2]
  long[speedPtr][3] := (globalY + globalX) '* 2
  long[speedPtr][1] := long[speedPtr][3]
      
PUB ComputeVector | localX, localY

  localX := F.Ffloat(-globalX) 
  localY := F.Ffloat(globalY)
   
  result := F.Atan2(localY, localX)

  localX := F.Fmul(localX, localX) 
  localY := F.Fmul(localY, localY)
  localX := F.Fadd(localX, localY) 
  localY := F.FSqr(localX)

  FComputeTargetSpeeds(localY, result)
  
PUB ComputeTargetSpeeds(localMagnitude, localAngle)

  localMagnitude := F.FFloat(localMagnitude)
  localMagnitude := F.FDiv(localMagnitude, fPseudoMultiplier)
  localAngle := F.FFloat(localAngle)
  localAngle := F.FDiv(localAngle, fPseudoMultiplier)
  localAngle := F.Radians(localAngle)
  FComputeTargetSpeeds(localMagnitude, localAngle)
  
PUB FComputeTargetSpeeds(fLocalMagnitude, fLocalAngle) | localIndex, fCosSpeed, fSinSpeed

  fLocalAngle := F.Fadd(fLocalAngle, R45)  
  'magnitude := F.FRound(fLocalMagnitude) 
  fSinSpeed := F.Sin(fLocalAngle)
  tempTargetSpeed[0] := F.FMul(fLocalMagnitude, fSinSpeed)
  tempTargetSpeed[0] := F.FRound(tempTargetSpeed[0])
  tempTargetSpeed[3] := tempTargetSpeed[0]
  fCosSpeed := F.Cos(fLocalAngle) 
  tempTargetSpeed[2] := F.FMul(fLocalMagnitude, fCosSpeed)
  tempTargetSpeed[2] := F.FRound(tempTargetSpeed[2])
  -tempTargetSpeed[2]
  tempTargetSpeed[1] := tempTargetSpeed[2]
    
  repeat localIndex from 0 to HIGHEST_WHEEL_ID
    if localIndex == 1 or localIndex == 3
      long[speedPtr][localIndex] := tempTargetSpeed[localIndex] + globalZ
    else
      long[speedPtr][localIndex] := tempTargetSpeed[localIndex] - globalZ
 