DAT programName         byte "FourMotorsAndEncoders130223a", 0     
CON
{{
  ******* Public Notes *******

  The object uses three cogs. One for the
  PWM motor control. One to read the four
  quadrature encoders. And a final cog to
  adjust the PWM based on the encoder readings
  on input received from the parent object.
  
}}
{
  ******* Private Notes *******

  130125a Use debug statements in parent object
  instead of child object.
  DemoFourMotorsAndEncoders130125a modified from
  DemoFourMotors130125a.
  DemoFourMotorsAndEncoders130125a add encoders.
  31c Comment out motor control.
  0201a Switch to Vex Treads robot.
  0203a Progress! Motor #0 at +1000 target speeds reads
  -945 speed from encoders.
  04a Move debug stuff to it's own cog
  04b Start adding speed feed back.
  06a Still trying to figure out PID stuff.
  So far there are these speed variables.
  bufferSpeed : speed from encoders, the highest this will likely get is 945 when
  motors are at full power
  targetSpeed : speed from input device (keyboard, joystick, etc.), highest is 1000 
  rampedSpeed : target speed limted by ramping method, highest 1000
  output : PWM value to send to motor, highest 1000
  pseudoRampedSpeed : rampedSpeed multiplied by pseudoreal multiplier to reduce
  rounding errors
  06d Compiles but doesn't really work. Speed doesn't appear to be
  passed to the child object. Encoder movement can set motor to full speed.
  07a Beautiful!
  07a Ramping isn't currently used becuase of the "MainLoop" method.
  07c Has big problems. The "Straight" method doesn't appear to work correctly.
  Motor #1 drives full speed and motor #0 isn't powered (at least not in a way to
  cause motion). Motor #0 may have both drive lines held high causing it to break
  (as in resist motion).
  08a Try a simplier ramping method with hardcoded values.
  08c Seems to work?
  09a Rover5 motor #0 spins all the time. The speed feedback appears to be
  reversed.
  11a Without acting on speed feedback, the motors appear to work correctly.
  11b Try to add feedback control.
  11e Finally seem to be getting PID to work.
  12a Seems to work.
  12b Try to improve ramping.
  12b Ramping seems to work okay.
  12c Try to improve P.
  12c Seems to work okay. I didn't add a mechanism to scale speeds if highest speed
  is above allowed max.
  12d Try adding turning.
  12d Turning seems to work but the treads come off the Rover 5.
  14a Save as FourMotorsAndEncoders130214a to begin making this a child object.
  14b Doesn't have a way to keep motors active. Still need to add continuous loop
  (this will likely cost another cog).
  17a Keep alive loop had been previously added.
  17a Try to add code to turn off proportional feedback in order to use
  this object to test encoder and motor directions.
  18a The keep alive loop hadn't been started in a new cog so control was
  never returned to the parent object.
  18a Add the cognew call.
  18b Add this object's ability to read encoder speed.
  18c Probably problem with ramping.
  22a Remove interrupt parameter from public motor control methods.
  22a Pretty big change.
  22a Clean up ramping. Private methods "StraightPrivate" and "TurnPrivate"
  are now only called once per coresponding public method calls.
  23a Change to use pointer to output speeds. Pointer will be passed
  to PWM object. This object will now receive a pointer to target speeds.
    
  ' Todo:
  Fix ramping so control returns to mainloop faster. (done)
  Fix timing in "Refresh" method.
  Eventually move the main loop to the motor control cog in order to
  free up a cog.
  
} 
  _CLKMODE = XTAL1 + PLL16X 
  _XINFREQ = 5_000_000
  
  'MAXSPEED = Pwm#PARENT_MAXSPEED
  PSEUDOREAL_MULTIPLIER = 100
  'PSEUDO_FULL_SPEED = FULL_SPEED * PSEUDOREAL_MULTIPLIER
  'DEFAULT_ACCELERATION = 8
  'PSEUDO_ACCELERATION = DEFAULT_ACCELERATION * PSEUDOREAL_MULTIPLIER
  'PARENT_MAXSPEED = Pwm#PARENT_MAXSPEED
  PID_INCREASE = 10
  PID_MULTIPLIER = PSEUDOREAL_MULTIPLIER * PID_INCREASE
  PSEUDOREAL_MAXSPEED = Pwm#PARENT_MAXSPEED * PSEUDOREAL_MULTIPLIER
  MOTORS_IN_USE = Encoders#TOTAL_ENCODERS
  HIGHEST_MOTOR_ID = MOTORS_IN_USE - 1

  MINIMUM_WAIT  = 3000 '1456

  SPEED_BUFFER_SIZE = Encoders#SPEED_BUFFER_SIZE
  ' state enumerate 
  #0, WAITING_FOR_MOTOR_ID, FIRST_DIGIT_SPEED, WAITING_FOR_SPEED

  ' speedSign enumeration
  '#0, POSITIVE, NEGATIVE

  ' oldDataType enumeration
  #0, NO_OLD_DATA, MOTOR_ONLY_OLD_DATA, MOTOR_AND_T_SPEED_OLD_DATA

  ' newDataType enumeration
  #0, NO_NEW_DATA, MOTOR_ONLY_NEW_DATA, WAITING_T_SPEED_NEW_DATA, PARTIAL_T_SPEED_NEW_DATA, {
    } FULL_T_SPEED_NEW_DATA, TIMED_ERROR_NEW_DATA

  ' motor names
  #0, FRONT_PORT, FRONT_STARBOARD, BACK_PORT, BACK_STARBOARD

  ' activePidConstant enumeration
  #0, PROPORTIONAL, INTEGRAL, DERIVATIVE, DELTA_TIME, TOP_SPEED_FEEDBACK, MAX_ACCELERATION

  ' command enumeration
  #0, NO_NEW_CMD, STRAIGHT_CMD, TURN_CMD, INDIVIDUAL_CMD
  
CON

  TITLE_X = 3                   ' location for title
  TITLE_Y = 0 '3
  DISPLAY_X = TITLE_X              
  DISPLAY_Y = TITLE_Y + 2 '3
  TARGET_SPEED_X = DISPLAY_X 
  TARGET_SPEED_Y = DISPLAY_Y 

  CURRENT_SPEED_X = TARGET_SPEED_X
  CURRENT_SPEED_Y = TARGET_SPEED_Y + 5 '6
  ENCODER_X = CURRENT_SPEED_X
  ENCODER_Y = CURRENT_SPEED_Y + 5 '6
  PIN_STATES_X = ENCODER_X
  PIN_STATES_Y = ENCODER_Y + 5 '6  
  OLD_INFO_X = PIN_STATES_X
  OLD_INFO_Y = PIN_STATES_Y + 3
  REQUEST_INPUT_X = OLD_INFO_X 
  REQUEST_INPUT_Y = OLD_INFO_Y + 2
  NEW_INFO_X = REQUEST_INPUT_X
  NEW_INFO_Y = REQUEST_INPUT_Y + 2

  SPEED_BUFFER_X = NEW_INFO_X 
  SPEED_BUFFER_Y = NEW_INFO_Y + 3
  PID_X = SPEED_BUFFER_X
  PID_Y = SPEED_BUFFER_Y + 6
  CONTROL_X = PID_X
  CONTROL_Y = PID_Y + 4 '5

  NEW_DATA_DISPLAY_TIME = 10    ' Keep this less than half the rollover time. 
  'PARAMETERS_TO_DISPLAY = 7

CON

  #0, NORMAL_OPERATION, TEST_DIRECTIONS_OPERATION
  
VAR

  'long enablePin[MOTORS_IN_USE], motorPinRed[MOTORS_IN_USE], motorPinBlack[MOTORS_IN_USE]
  long loopStack[100]
  'long previousTSpeed[MOTORS_IN_USE], frozenTSpeed[MOTORS_IN_USE * 2]
  'long encoder[MOTORS_IN_USE]
  'long bufferSpeed[MOTORS_IN_USE]
  'long tempSpeed, speedSign, previousTargetSpeed[MOTORS_IN_USE], previousMotorIndex
  'long newDataTimer, newDataInterval                                                            
  'long frozenCSpeed[MOTORS_IN_USE * 2], previousCSpeed[MOTORS_IN_USE]
  'long frozenEncoder[MOTORS_IN_USE], previousEncoder[MOTORS_IN_USE]
  'long previousIna, frozenIna, frozenBuffer[Encoders#SPEED_BUFFER_SIZE * Encoders#TOTAL_ENCODERS]
  long nextRefreshTime, refreshInterval
  'long pseudoIntegral[MOTORS_IN_USE], pseudoDerivative[MOTORS_IN_USE]
  'long previousPseudoError[MOTORS_IN_USE], previousError[MOTORS_IN_USE]
  long loopCount', targetSpeed[MOTORS_IN_USE]', rampedSpeed[MOTORS_IN_USE]
  'long pseudoAccelD[MOTORS_IN_USE]
 
  long previousGlobalTime, globalTime
  long pRampingSteps, globalRampingSteps', debugCount, uSecond
  'long rawOutput[MOTORS_IN_USE]', pDebubFlag, debugFlag
  long pseudoTargetSpeed[MOTORS_IN_USE]', globalIndex
  long directionCorrection[MOTORS_IN_USE]

  long kp, fastestPossible, pseudoMaxAcceleration
  long command, globalSpeed, globalTurnAmount', interrupt
  long globalIndexPtr, encoderSpeedPtr, targetPtr

  ' keep below in order and together
  long arrayPtr, speedPtr, bufferPtr
  long output[MOTORS_IN_USE], pseudoError[MOTORS_IN_USE]
  long returnedPtr, pseudoRampedSpeed[MOTORS_IN_USE]  
  ' keep above in order and together

  byte motorIndex, oldDataType, newDataType, rampFlag
  byte activePidConstant', keyboardInputFlag, rampingInProgress
  'byte testDirectionsFlag
   
OBJ

  Pwm : "FourMotors130223a"
  Encoders : "FourQuadratureEncoders130215a"
   
PUB Start(enablePtr, positivePinPtr, negativePinPtr, directionPtr, {
  } refreshInterval_, proportionalConstant, fastestPossible_, {
  } pseudoMaxAcceleration_, globalIndexPtr_, {
  } encoderPinPtr, encoderPtr, speedPtr_, targetPtr_)

  longmove(@directionCorrection, directionPtr, MOTORS_IN_USE) 
  
  rampFlag := 1
  'uSecond := clkfreq / 1_000_000
  refreshInterval := refreshInterval_
  kp := proportionalConstant
  fastestPossible := fastestPossible_
  pseudoMaxAcceleration := pseudoMaxAcceleration_
  globalIndexPtr := globalIndexPtr_
  'testDirectionsFlag := testDirectionsFlag_
  encoderSpeedPtr := speedPtr_
  targetPtr := targetPtr_
  
  Encoders.StartEncoders(encoderPinPtr, encoderPtr, speedPtr_)
  
  Pwm.Start(enablePtr, positivePinPtr, negativePinPtr, @output)
  
  arrayPtr := Pwm.GetArrayLocation
  speedPtr := Pwm.GetSpeedLocation
  bufferPtr := Encoders.GetBufferLocation
  result := @arrayPtr
  'targetSpeed
  nextRefreshTime := cnt + refreshInterval

  cognew(MainLoop, @loopStack) 
  
PRI MainLoop
'' This loop runs in its own cog. Control of object is done
'' by setting the variable "command" through public methods.
  repeat
    case command
      STRAIGHT_CMD:
        StraightPrivate
      TURN_CMD:
        TurnPrivate
      INDIVIDUAL_CMD:
        SetSpeedsPrivate
      'other: 'NO_NEW_CMD
        
PUB Straight(localSpeed)

  globalSpeed := localSpeed
  command := STRAIGHT_CMD
  
  'result := @targetSpeed
    
PUB Turn(localSpeed, turnAmount)
'' A positive turn is counter clockwise.

  globalSpeed := localSpeed
  globalTurnAmount := turnAmount
  command := TURN_CMD

  'result := @targetSpeed
  
PUB RefreshSpeeds'(targetPtr)

  'longmove(@targetSpeed, targetPtr, MOTORS_IN_USE)
  command := INDIVIDUAL_CMD
   
PUB SetSpeed(localMotorId, localSpeed) 

  long[targetPtr][localMotorId] := localSpeed
  command := INDIVIDUAL_CMD

  
PRI StraightPrivate

  repeat result from 0 to HIGHEST_MOTOR_ID
    long[targetPtr][result] := globalSpeed
  SetSpeedsPrivate
  command := INDIVIDUAL_CMD ' since target speeds have now been set
  
PRI TurnPrivate | localIndex
'' A positive turn is counter clockwise.
  
  repeat result from 0 to constant((MOTORS_IN_USE / 2) - 1)
    long[targetPtr][result * 2] := globalSpeed - globalTurnAmount
    long[targetPtr][(result * 2) + 1] := globalSpeed + globalTurnAmount
  SetSpeedsPrivate
  command := INDIVIDUAL_CMD
  
PRI SetSpeedsPrivate | localIndex, pseudoMaxSoFar, pseudoLocalAccel[4] 'MOTORS_IN_USE]
'' This is the main method used to control the
'' motors.

  pseudoMaxSoFar := 0
  repeat result from 0 to HIGHEST_MOTOR_ID
    pseudoTargetSpeed[result] := directionCorrection[result] * {
    } long[targetPtr][result] * PSEUDOREAL_MULTIPLIER
 
    pseudoLocalAccel[result] := pseudoTargetSpeed[result] - pseudoRampedSpeed[result]
    if ||pseudoLocalAccel[result] > pseudoMaxSoFar
      pseudoMaxSoFar := ||pseudoLocalAccel[result]

  if pseudoMaxSoFar > pseudoMaxAcceleration
    ' Ramp all motors proportionally the same.
    ' Motors that aren't being accelerated too fast will still get ramped
    ' so the proportional speeds will be the accurate.
    repeat result from 0 to HIGHEST_MOTOR_ID 
      pseudoLocalAccel[result] *= pseudoMaxAcceleration
      pseudoLocalAccel[result] /= pseudoMaxSoFar
      pseudoRampedSpeed[result] += pseudoLocalAccel[result]
  else
    ' All target speeds may be reached in one step.
    repeat result from 0 to HIGHEST_MOTOR_ID
      pseudoRampedSpeed[result] := pseudoTargetSpeed[result]

 
  Refresh

PRI Refresh 
 
  ComputeP

  ' ** should I check to see if any outputs have changed?
  ' if changeFlag?
  returnedPtr := Pwm.Update

  if long[globalIndexPtr] > negx ' keep globalIndex from rolling over to positive                        
    long[globalIndexPtr]--
  previousGlobalTime := globalTime
  globalTime := nextRefreshTime - cnt   ' **** this could be simplified
  if globalTime > MINIMUM_WAIT
    waitcnt(nextRefreshTime)
    nextRefreshTime += refreshInterval
  else
    nextRefreshTime := cnt + refreshInterval
    
PUB ComputeP
' Only the proportional algorithm is used in this method.
' Possibly add integral and derivative

  repeat result from 0 to HIGHEST_MOTOR_ID
    pseudoError[result] := pseudoRampedSpeed[result] - {
    } ((long[encoderSpeedPtr][result] * PSEUDOREAL_MAXSPEED) / fastestPossible)
     
    output[result] := (pseudoRampedSpeed[result] + ((kp * pseudoError[result] ) {
    } / PID_INCREASE)) / PSEUDOREAL_MULTIPLIER

PUB GetArrayLocation

  result := Pwm.GetArrayLocation

PUB GetSpeedLocation

  result := Pwm.GetSpeedLocation

PUB GetFileName
'' This method returns the name of this object.
'' This method is not needed for proper operation of object
'' and may be deleted.

  result := @programName

PUB GetPwmGrandChildName
'' This method returns the name of this object.
'' This method is not needed for proper operation of object
'' and may be deleted.

  result := Pwm.GetFileName

PUB GetEncoderGrandChildName
'' This method returns the name of this object.
'' This method is not needed for proper operation of object
'' and may be deleted.

  result := Encoders.GetFileName

PUB GetBufferLocation

  result := Encoders.GetBufferLocation

PUB ChangeProportionalConstant(value)

  result := kp
  result += value
  result #>= 0
  kp := result

{PUB TestMotors(localPtr)

  if testDirectionsFlag == NORMAL_OPERATION
    SetSpeeds(localPtr, false)
  else  
    repeat result from 0 to HIGHEST_MOTOR_ID
      output[result] := long[localPtr][result]    
    result := Pwm.Update(output[0], output[1], output[2], output[3])
   
PUB TestMotor(motorId, localSpeed)

  if testDirectionsFlag == NORMAL_OPERATION
    SetSpeed(motorIndex, localSpeed, false)
  else  
    Pwm.UpdateSingle(motorId, localSpeed)

}   
DAT '' License
{{
*********************************************************************************************
*                             TERMS OF USE: MIT License                                     *                                                                   
*********************************************************************************************
* Permission is hereby granted, free of charge, to any person obtaining a copy of this      *
* software and associated documentation files (the "Software"), to deal in the Software     *
* without restriction, including without limitation the rights to use, copy, modify,        *
* merge, publish, distribute, sublicense, and/or sell copies of the Software, and to        *
* permit persons to whom the Software is furnished to do so, subject to the following       *
* conditions:                                                                               *
*                                                                                           *                                    
* The above copyright notice and this permission notice shall be included in all copies or  *
* substantial portions of the Software.                                                     *
*                                                                                           *                                   
* The software is provided "as is", without warranty of any kind, express or implied,       *
* including but not limited to the warranties of merchantability, fitness for a particular  *
* and noninfringment. In no event shall the authors or copyright holders be liable for any  *
* claim, damages or other liability, whether in an action of contract, tort or otherwise,   *
* arising from, of or in connection with the software of the use of other dealings in the   *
* software.                                                                                 *
*********************************************************************************************
}}            